@@ -1,5 +1,4 @@
 import logging
-import os
 import time
 import threading
 import math
@@ -11,6 +10,7 @@ import itertools
 import hashlib
 import weakref
 import queue
+import pickle
 from collections import defaultdict
 from collections.abc import MutableMapping
 
@@ -80,6 +80,9 @@ LOG_INVALID_DEBUG_LEVEL_MSG = "ERR Invalid debug level."
 LUA_COMMAND_ARG_MSG = "ERR Lua redis() command arguments must be strings or integers"
 LUA_WRONG_NUMBER_ARGS_MSG = "ERR wrong number or type of arguments"
 SCRIPT_ERROR_MSG = "ERR Error running script (call to f_{}): @user_script:?: {}"
+RESTORE_KEY_EXISTS = "BUSYKEY Target key name already exists."
+RESTORE_INVALID_CHECKSUM_MSG = "ERR DUMP payload version or checksum are wrong"
+RESTORE_INVALID_TTL_MSG = "ERR Invalid TTL value, must be >= 0"
 
 FLAG_NO_SCRIPT = 's'      # Command not allowed in scripts
 
@@ -811,6 +814,7 @@ class FakeSocket:
     def _process_command(self, fields):
         if not fields:
             return
+        func_name = None
         try:
             func, func_name = self._name_to_func(fields[0])
             sig = func._fakeredis_sig
@@ -831,6 +835,11 @@ class FakeSocket:
                 # TODO: should not apply if the exception is from _run_command
                 # e.g. watch inside multi
                 self._transaction_failed = True
+            if func_name == 'exec' and exc.value.startswith('ERR '):
+                exc.value = 'EXECABORT Transaction discarded because of: ' + exc.value[4:]
+                self._transaction = None
+                self._transaction_failed = False
+                self._clear_watches()
             result = exc
         result = self._decode_result(result)
         if not isinstance(result, NoResponse):
@@ -879,6 +888,7 @@ class FakeSocket:
         returned exactly once.
         """
         pattern = None
+        type = None
         count = 10
         if len(args) % 2 != 0:
             raise SimpleError(SYNTAX_ERROR_MSG)
@@ -889,6 +899,8 @@ class FakeSocket:
                 count = Int.decode(args[i + 1])
                 if count <= 0:
                     raise SimpleError(SYNTAX_ERROR_MSG)
+            elif casematch(args[i], b'type'):
+                type = args[i + 1]
             else:
                 raise SimpleError(SYNTAX_ERROR_MSG)
 
@@ -898,11 +910,20 @@ class FakeSocket:
         result_cursor = cursor + count
         result_data = []
 
-        if pattern is not None:
-            regex = compile_pattern(pattern)
+        regex = compile_pattern(pattern) if pattern is not None else None
+
+        def match_key(key):
+            return regex.match(key) if pattern is not None else True
+
+        def match_type(key):
+            if type is not None:
+                return casematch(self.type(self._db[key]).value, type)
+            return True
+
+        if pattern is not None or type is not None:
             for val in itertools.islice(data, cursor, result_cursor):
                 compare_val = val[0] if isinstance(val, tuple) else val
-                if regex.match(compare_val):
+                if match_key(compare_val) and match_type(compare_val):
                     result_data.append(val)
         else:
             result_data = data[cursor:result_cursor]
@@ -1209,6 +1230,37 @@ class FakeSocket:
         else:
             return out
 
+    @command((Key(missing_return=None),))
+    def dump(self, key):
+        value = pickle.dumps(key.value)
+        checksum = hashlib.sha1(value).digest()
+        return checksum + value
+
+    @command((Key(), Int, bytes), (bytes,))
+    def restore(self, key, ttl, value, *args):
+        replace = False
+        i = 0
+        while i < len(args):
+            if casematch(args[i], b'replace'):
+                replace = True
+                i += 1
+            else:
+                raise SimpleError(SYNTAX_ERROR_MSG)
+        if key and not replace:
+            raise SimpleError(RESTORE_KEY_EXISTS)
+        checksum, value = value[:20], value[20:]
+        if hashlib.sha1(value).digest() != checksum:
+            raise SimpleError(RESTORE_INVALID_CHECKSUM_MSG)
+        if ttl < 0:
+            raise SimpleError(RESTORE_INVALID_TTL_MSG)
+        if ttl == 0:
+            expireat = None
+        else:
+            expireat = self._db.time + ttl / 1000.0
+        key.value = pickle.loads(value)
+        key.expireat = expireat
+        return OK
+
     # Transaction commands
 
     def _clear_watches(self):
@@ -1240,6 +1292,7 @@ class FakeSocket:
             raise SimpleError(WITHOUT_MULTI_MSG.format('EXEC'))
         if self._transaction_failed:
             self._transaction = None
+            self._clear_watches()
             raise SimpleError(EXECABORT_MSG)
         transaction = self._transaction
         self._transaction = None
@@ -2601,39 +2654,9 @@ class FakeSelector(BaseSelector):
 class FakeConnection(redis.Connection):
     description_format = "FakeConnection<db=%(db)s>"
 
-    def __init__(self, server, db=0, username=None, password=None,
-                 socket_timeout=None, socket_connect_timeout=None,
-                 socket_keepalive=False, socket_keepalive_options=None,
-                 socket_type=0, retry_on_timeout=False,
-                 encoding='utf-8', encoding_errors='strict',
-                 decode_responses=False, parser_class=_DummyParser,
-                 socket_read_size=65536, health_check_interval=0,
-                 client_name=None):
-        self.pid = os.getpid()
-        self.db = db
-        self.username = username
-        self.client_name = client_name
-        self.password = password
-        # Allow socket attributes to be passed in and saved even if they aren't used
-        self.socket_timeout = socket_timeout
-        self.socket_connect_timeout = socket_connect_timeout or socket_timeout
-        self.socket_keepalive = socket_keepalive
-        self.socket_keepalive_options = socket_keepalive_options or {}
-        self.socket_type = socket_type
-        self.retry_on_timeout = retry_on_timeout
-        self.encoder = redis.connection.Encoder(encoding, encoding_errors, decode_responses)
-        self._description_args = {'db': self.db}
-        self._connect_callbacks = []
-        self._buffer_cutoff = 6000
-        self._server = server
-        # self._parser isn't used for anything, but some of the
-        # base class methods depend on it and it's easier not to
-        # override them.
-        self._parser = parser_class(socket_read_size=socket_read_size)
-        self._sock = None
-        # added in redis==3.3.0
-        self.health_check_interval = health_check_interval
-        self.next_health_check = 0
+    def __init__(self, *args, **kwargs):
+        self._server = kwargs.pop('server')
+        super().__init__(*args, **kwargs)
 
     def connect(self):
         super().connect()
